import enum
from itertools import zip_longest

from destruct import parse, Type, Struct
from .util import MultiEnum
from .common import CLICodedToken, CLITableType, CLIElementType


class CLICompressedInt(Type):
    def __init__(self, signed=True):
        self.signed = signed

    def parse(self, input, context):
        val = input.read(1)[0]
        if (val >> 5) == 0b110:
            nbytes = 4
            b = bytearray([val & 0b11111])
        elif (val >> 6) == 0b10:
            nbytes = 2
            b = bytearray([val & 0b111111])
        else:
            nbytes = 1
            b = bytearray([val])
        b.extend(input.read(nbytes - 1))
        return int.from_bytes(b, byteorder='big', signed=self.signed)

class CLICompressedUInt(Type):
    def __new__(self):
        return CLICompressedInt(signed=False)


class CLINativeElementType(enum.Enum):
    Boolean  = 0x2
    I1       = 0x3
    U1       = 0x4
    I2       = 0x5
    U2       = 0x6
    I4       = 0x7
    U4       = 0x8
    I8       = 0x9
    U8       = 0xA
    R4       = 0xB
    R8       = 0xC
    LPStr    = 0x14
    LPWStr   = 0x15
    Int      = 0x1F
    UInt     = 0x20
    Function = 0x26
    Array    = 0x2A

class CLINativeType:
    pass

class CLINativeIntrinsicType(CLINativeType):
    def __init__(self, type: CLIElementType):
        self.type = type

    def __str__(self):
        return self.type.name.lower()

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.type.name)

class CLINativeArrayType(CLINativeType):
    def __init__(self, type, count, count_param):
        self.type = type
        self.count = count
        self.count_param = count_param

    def __str__(self):
        return str(self.type) + '[{}]'.format(self.count if self.count is not None else '')

    def __repr__(self):
        return '<{}: {}[{}] @ {}>'.format(
            self.__class__.__name__, self.type.name,
            self.count if self.count is not None else '', self.count_param
        )

class CLINativeSignature(Struct, nocopy=True):
    type        = Enum(CLINativeElementType, CLICompressedUInt())
    elem_type   = Maybe(Enum(CLINativeElementType, CLICompressedUInt()))
    count_param = Maybe(CLICompressedUInt())
    count       = Maybe(CLICompressedUInt())

    def parse(self, input, context):
        val = super().parse(input, context)
        if val.type == CLINativeElementType.Array:
            return CLINativeArrayType(val.elem_type, val.count, val.count_param)
        return CLINativeIntrinsicType(val.type)


class CLIType:
    pass

class CLIPrimitiveType(CLIType):
    def __init__(self, type: CLIElementType):
        self.type = type

    def __str__(self):
        return self.type.name.lower()

    def __repr__(self):
        return '<{}: {}>'.format(self.__class__.__name__, self.type.name)

class CLIPointerType(CLIType):
    def __init__(self, child: CLIType):
        self.child = child

    def __str__(self):
        return str(self.child) + '*'

    def __repr__(self):
        return '<{}: {!r}>'.format(self.__class__.__name__, self.child)

class CLIRefType(CLIType):
    def __init__(self, child: CLIType):
        self.child = child

    def __str__(self):
        return str(self.child) + '&'

    def __repr__(self):
        return '<{}: {!r}>'.format(self.__class__.__name__, self.child)

class CLIPinnedType(CLIType):
    def __init__(self, child: CLIType):
        self.child = child

    def __str__(self):
        return str(self.child) + ' fixed&'

    def __repr__(self):
        return '<{}: {!r}>'.format(self.__class__.__name__, self.child)

class CLIUserTypeKind(enum.Enum):
    Class     = 'class'
    ValueType = 'enum'

class CLIUserType(CLIType):
    def __init__(self, token, kind):
        self.token = token
        self.kind = kind

    def __str__(self):
        if self.token.table == CLITableType.TypeRef:
            prefix = 'ext-'
        else:
            prefix = ''
        return '{}{}#{}'.format(prefix, self.kind.value, self.token.row)

    def __repr__(self):
        return '<{} ({}): {!r}>'.format(self.__class__.__name__, self.kind.value, self.token)

class CLIOptionalType(CLIType):
    def __init__(self, child):
        self.child = child

    def __str__(self):
        return str(self.child) + '?'

    def __repr__(self):
        return '<{}: {!r}>'.format(self.__class__.__name__, self.child)

class CLIRequiredType(CLIType):
    def __init__(self, child):
        self.child = child

    def __str__(self):
        return str(self.child) + '!'

    def __repr__(self):
        return '<{}: {!r}>'.format(self.__class__.__name__, self.child)

class CLIGenericParamScope(enum.Enum):
    Type   = 'type'
    Method = 'method'

class CLIGenericParamType(CLIType):
    def __init__(self, index, scope):
        self.index = index
        self.scope = scope

    def __str__(self):
        return '#' + str(self.index)

    def __repr__(self):
        return '<{}: {} @ {}>'.format(self.__class__.__name__, self.index, self.scope.value)

class CLIGenericInstantiationType(CLIType):
    def __init__(self, child, args):
        self.child = child
        self.args = args

    def __str__(self):
        return str(self.child) + '[' + ',  '.join(str(a) for a in self.args) + ']'

    def __repr__(self):
        return '<{}: {!r}[{}]>'.format(self.__class__.__name__, self.child, ',  '.join(repr(a) for a in self.args))

class CLIArrayType(CLIType):
    def __init__(self, child, rank, sizes=None, lowers=None):
        self.child = child
        self.rank = rank
        self.sizes = sizes or []
        self.lowers = lowers or []

    def __str__(self):
        return str(self.child) + ''.join(
            '[{}{}]'.format(str(lower) + '...' if lower else '', upper + (lower or 0) if upper else '')
            for _, lower, upper in zip_longest(range(self.rank), self.lowers, self.sizes)
        )

    def __repr__(self):
        return '<{}: {!r}{}>'.format(self.__class__.__name__, self.child, ''.join(
            '[{}{}]'.format(str(lower) + '...' if lower else '', upper + (lower or 0) if upper else '')
            for _, lower, upper in zip_longest(range(self.rank), self.lowers, self.sizes)
        ))

class CLIElement(Type):
    def __init__(self, child):
        self.child = child

    def parse(self, input, context):
        nested_types = {
            CLIElementType.Pointer: CLIPointerType,
            CLIElementType.ByRef:   CLIRefType,
            CLIElementType.Pinned:  CLIPinnedType,
        }
        token_types = {
            CLIElementType.Class:       lambda t: CLIUserType(t, CLIUserTypeKind.Class),
            CLIElementType.ValueType:   lambda t: CLIUserType(t, CLIUserTypeKind.ValueType),
            CLIElementType.OptModifier: CLIOptionalType,
            CLIElementType.ReqModifier: CLIRequiredType,
        }
        int_types = {
            CLIElementType.Var:  lambda i: CLIGenericParamType(i, CLIGenericParamScope.Type),
            CLIElementType.MVar: lambda i: CLIGenericParamType(i, CLIGenericParamScope.Method),
        }
        val = CLIElementType(parse(self.child, input, context))
        if val in nested_types:
            with context.enter(val.name, self):
                val = nested_types[val](self.parse(input, context))
        elif val in token_types:
            with context.enter(val.name, self.child):
                raw = parse(self.child, input, context)
                token = parse(CLICodedToken([CLITableType.TypeDef, CLITableType.TypeRef]), raw.to_bytes(4, 'little'), context)
                val = token_types[val](token)
        elif val in int_types:
            with context.enter(val.name, self.child):
                val = int_types[val](parse(self.child, input, context))
        elif val == CLIElementType.Array:
            with context.enter(val.name, self):
                type = self.parse(input, context)
                rank = parse(self.child, input, context)
                nbounds = parse(self.child, input, context)
                bounds = [parse(self.child, input, context) for _ in range(nbounds)]
                nlo = parse(self.child, input, context)
                lo  = [parse(self.child, input, context) for _ in range(nlo)]
            val = CLIArrayType(type, rank, bounds, lo)
        elif val == CLIElementType.GenericInst:
            with context.enter(val.name, self):
                type = self.parse(input, context)
                nargs = parse(self.child, input, context)
                args = []
                for i in range(nargs):
                    with context.enter(i, self):
                        args.append(self.parse(input, context))
            val = CLIGenericInstantiationType(type, args)
        else:
            val = CLIPrimitiveType(val)
        return val


class CLISignatureType(enum.Enum):
    Default  = 0x0
    C        = 0x1
    StdCall  = 0x2
    ThisCall = 0x3
    FastCall = 0x4
    VarArg   = 0x5
    Field    = 0x6
    LocalVar = 0x7
    Property = 0x8

class CLISignatureFlags(enum.Flag):
    Generic      = 0x10
    HasThis      = 0x20
    ExplicitThis = 0x40

class CLISignatureAttributes(MultiEnum):
    type  = (CLISignatureType,  0x0F)
    flags = (CLISignatureFlags, 0xF0)

class CLIMethodSignature(Struct, nocopy=True):
    param_count = CLICompressedUInt()
    ret_type    = CLIElement(CLICompressedUInt())
    params      = Arr(CLIElement(CLICompressedUInt()))

    def on_param_count(self, spec, context):
        spec.params.count = self.param_count

    def __str__(self):
        return '(' + ', '.join(str(p) for p in self.params) + ') -> ' + str(self.ret_type)

class CLIFieldSignature(Struct, nocopy=True):
    type = CLIElement(CLICompressedUInt())

    def __str__(self):
        return str(self.type)

class CLIPropertySignature(Struct, nocopy=True):
    param_count = CLICompressedUInt()
    type        = CLIElement(CLICompressedUInt())
    params      = Arr(CLIElement(CLICompressedUInt()))

    def on_param_count(self, spec, context):
        spec.params.count = self.param_count

    def __str__(self):
        return '(' + ', '.join(str(p) for p in self.params) + ') -> ' + str(self.type)

class CLILocalVarSignature(Struct, nocopy=True):
    count = CLICompressedUInt()
    vars  = Arr(CLIElement(CLICompressedUInt()))

    def on_count(self, spec, context):
        spec.vars.count = self.count

class CLISignature(Struct, nocopy=True):
    attribs   = Enum(CLISignatureAttributes, UInt(8))
    signature = Switch(options={
        CLISignatureType.Default:  CLIMethodSignature,
        CLISignatureType.C:        CLIMethodSignature,
        CLISignatureType.StdCall:  CLIMethodSignature,
        CLISignatureType.ThisCall: CLIMethodSignature,
        CLISignatureType.FastCall: CLIMethodSignature,
        CLISignatureType.VarArg:   CLIMethodSignature,
        CLISignatureType.Field:    CLIFieldSignature,
        CLISignatureType.Property: CLIPropertySignature,
        CLISignatureType.LocalVar: CLILocalVarSignature,
    })

    def on_attribs(self, spec, context):
        spec.signature.selector = self.attribs.type

    def __str__(self):
        return str(self.signature)
